




from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow.step.inputs import (
    ConnectionInputValue, 
    InputValue, 
    WorkflowInputInputValue
)


def update_components_array(value: InputValue, janis: Workflow) -> None:
    """
    correct information about tool components using InputValue.
    ie if workflow input is array, but tool component is not, 
    update tool component to be an array. 
    """
    pass
    # if value.component:  # if linked to tool component
    #     if isinstance(value, WorkflowInputInputValue):
    #         if not value.is_runtime:
    #             w_inp = janis.get_input(query_uuid=value.input_uuid)
    #             value.component.forced_array = w_inp.array
    #     elif isinstance(value, ConnectionInputValue):
    #         s_out = janis.get_step_output(query_uuid=value.out_uuid) # type: ignore
    #         value.component.forced_array = s_out.tool_output.array






from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Optional
from uuid import uuid4
from enum import Enum, auto
from Cheetah.Template import Template
from galaxy.util import unicodify

from .. import utils
from galaxy2janis.gx.command.cmdstr import constructs

from galaxy2janis import expressions
from galaxy2janis.expressions.patterns import CHEETAH_EDGE_CASE_INPUT


def get_blocks(ptr: int, lines: list[str], indent_level: int) -> list[CheetahBlock]:
    block_lines = LineFactory(lines).get_lines()
    factory = BlockFactory(ptr, block_lines, indent_level)
    return factory.get_blocks()

def get_next_block(ptr: int, lines: list[str]) -> CheetahBlock:
    blocks = get_blocks(ptr=ptr, lines=lines[ptr:], indent_level=0)
    return blocks[0]


@dataclass
class BlockLine:
    line_num: int
    indent: int
    text: str

class LineFactory:
    """creates BlockLines from normal text lines"""

    def __init__(self, lines: list[str]):
        self.text_lines = lines
        self.block_lines: list[BlockLine] = []
        self.construct_tracker: constructs.ConstructTracker = constructs.ConstructTracker()

    def get_lines(self) -> list[BlockLine]:
        if len(self.text_lines) == 0:
            return []
        else:
            for i, line in enumerate(self.text_lines):
                self.add_block_line(i, line)
            return self.block_lines

    def add_block_line(self, i: int, line: str) -> None:
        indent = self.construct_tracker.stack.depth
        self.construct_tracker.update(line) # LEAVE THIS HERE
        block_line = BlockLine(
            line_num=i,
            indent=indent,
            text=line
        )
        self.block_lines.append(block_line)
    
    


IGNORE_CONSTRUCTS = constructs.CH_WITHIN_CONDITIONAL | constructs.CH_CLOSE_CONDITIONAL | constructs.CH_CLOSE_LOOP | constructs.CH_CLOSE_FUNC

class BlockFactory:
    """
    gets all cheetah blocks occurring in self.lines on a particular 
    indent level
    """
    def __init__(self, offset: int, block_lines: list[BlockLine], indent_level: int):
        self.offset = offset
        self.block_lines = block_lines
        self.target_indent = indent_level
        self.active_lines: list[BlockLine] = []
        self.blocks: list[CheetahBlock] = []

    def get_blocks(self) -> list[CheetahBlock]:
        for line in self.block_lines:
            if self.start_block(line): # new block begins
                self.add_block() # add previous block
                self.active_lines = [] # reset active lines
            if self.within_block(line):
                self.active_lines.append(line)
        self.add_block()
        self.filter_blocks()
        return self.blocks

    def start_block(self, line: BlockLine) -> bool:
        if line.indent == self.target_indent:
            return True
        return False
    
    def within_block(self, line: BlockLine) -> bool:
        if line.indent >= self.target_indent:
            return True
        return False

    def add_block(self) -> None:
        if self.active_lines:
            block = CheetahBlock(
                type=self.select_block_type(),
                start=self.active_lines[0].line_num + self.offset,
                stop=self.active_lines[-1].line_num + self.offset,
                lines=[ln.text for ln in self.active_lines]
            )
            self.blocks.append(block)
    
    def select_block_type(self) -> BlockType:
        line = self.active_lines[0].text
        types = [
            (constructs.CH_OPEN_CONDITIONAL, BlockType.CONDITIONAL),
            (constructs.CH_OPEN_LOOP, BlockType.LOOP),
            (constructs.CH_OPEN_FUNC, BlockType.FUNCTION),
            (constructs.CH_ENV, BlockType.INLINE_CH),
            (constructs.LINUX_ALIAS, BlockType.INLINE_ALIAS),
        ]
        for construct, block_type in types:
            if any ([line.startswith(text) for text in construct]):
                return block_type
        return BlockType.INLINE

    def filter_blocks(self):
        filtered: list[CheetahBlock] = []
        for block in self.blocks:
            if len(block.lines) == 1:
                line = block.lines[0]
                if not any ([line.startswith(text) for text in IGNORE_CONSTRUCTS]):
                    filtered.append(block)
            else:
                filtered.append(block)
        self.blocks = filtered



class EvaluationStrategy(ABC):
    def __init__(self, lines: list[str], input_dict: dict[str, Any]):
        self.lines = lines
        self.input_dict = input_dict
    
    def eval(self) -> Optional[list[str]]:
        template = self.prepare_template()
        outcome = self.evaluate_template(template)
        if outcome is not None: 
            return self.handle_outcome(outcome)
        return None

    @abstractmethod
    def prepare_template(self) -> list[str]:
        """prepares the template text for evaluation"""
        ...
    
    def evaluate_template(self, source_lines: list[str]) -> Optional[list[str]]:
        """performs cheetah evaluation of template"""
        try:
            source = utils.join_lines(source_lines)
            t = Template(source, searchList=[self.input_dict]) # type: ignore
            evaluation = str(unicodify(t))
            return utils.split_lines_blanklines(evaluation)
        except:
            return None

    @abstractmethod
    def handle_outcome(self, outcome: list[str]) -> list[str]:
        """handles the evaluated text (if successful) and applies any transformations needed"""
        ...



class InlineEvaluationStrategy(EvaluationStrategy):

    def prepare_template(self) -> list[str]:
        assert(len(self.lines) == 1)
        return self.lines

    def handle_outcome(self, outcome: list[str]) -> list[str]:
        return outcome


class ConditionalEvaluationStrategy(EvaluationStrategy):

    def prepare_template(self) -> list[str]:
        self.mask_children()
        return self.lines

    def handle_outcome(self, outcome: list[str]) -> list[str]:
        output = self.create_blank_output()
        output = self.restore_surviving_children(outcome, output)
        return output

    def mask_children(self) -> None:
        """prepares text ready for evaluation. swaps child blocks with identifiers"""
        self.masked_blocks: dict[str, CheetahBlock] = {}
        for child in self.get_child_blocks():
            self.masked_blocks[child.uuid] = child  # register identifier/block
            self.substitute_identifier(child)

    def get_child_blocks(self) -> list[CheetahBlock]:
        """returns next level of cheetah blocks within this block"""
        return get_blocks(
            ptr=0,
            lines=self.lines, 
            indent_level=1
        )

    def substitute_identifier(self, block: CheetahBlock) -> None:
        """swaps block block with identifier in template"""
        old_lines_top = self.lines[:block.start]
        old_lines_bottom = self.lines[block.stop+1:]
        self.lines = old_lines_top + [block.uuid] * block.height + old_lines_bottom

    def create_blank_output(self) -> list[str]:
        return [''] * len(self.lines)

    def restore_surviving_children(self, evaluation: list[str], output: list[str]) -> list[str]:
        surviving_children = self.get_surviving_children(evaluation)
        for identifier, block in surviving_children.items():
            line_num = self.get_identifier_line_num(self.lines, identifier)
            if line_num is not None:
                block = self.masked_blocks[identifier]
                output = self.substitute_block(output, block)
        return output

    def get_surviving_children(self, evaluation: list[str]) -> dict[str, Any]:
        out: dict[str, Any] = {}
        for identifier, block in self.masked_blocks.items():
            position = self.get_identifier_line_num(evaluation, identifier)
            if position is not None:
                out[identifier] = block
        return out

    def get_identifier_line_num(self, lines: list[str], identifier: str) -> Optional[int]:
        for i, line in enumerate(lines):
            if line == identifier:
                return i
        return None

    def substitute_block(self, output: list[str], block: CheetahBlock) -> list[str]:
        """swaps block block with identifier in template"""
        old_lines_top = output[:block.start]
        old_lines_bottom = output[block.stop+1:]
        return old_lines_top + block.lines + old_lines_bottom



class BlockType(Enum):
    MAIN            = auto()
    INLINE          = auto()
    INLINE_CH       = auto()
    INLINE_ALIAS    = auto()
    CONDITIONAL     = auto()
    FUNCTION        = auto()
    LOOP            = auto()


class CheetahBlock:
    """
    The smallest unit of evaluatable cheetah logic
    examples: single line statement, conditional block, loop block
    """
    def __init__(self, type: BlockType, start: int, stop: int, lines: list[str]):
        self.type = type
        self.start = start
        self.stop = stop
        self.lines = lines
        self.uuid: str = str(uuid4())
        self.evaluated: bool = False
    
    @property 
    def height(self) -> int:
        return self.stop - self.start + 1
    
    def evaluate(self, input_dict: dict[str, Any]) -> None:
        if self.should_evaluate(input_dict):
            strategy = self.get_eval_strategy(input_dict)
            evaluation = strategy.eval()
            if evaluation is not None:
                assert(len(self.lines) == len(evaluation))
                self.evaluated = True
                self.lines = evaluation
        
    def should_evaluate(self, input_dict: dict[str, Any]) -> bool:
        """dictates whether this block should be evaluated or left as original text"""
        permitted_blocks = [BlockType.INLINE, BlockType.INLINE_ALIAS, BlockType.INLINE_CH, BlockType.CONDITIONAL]
        if self.lines == ['']:
            return False
        elif self.type not in permitted_blocks:
            return False
        elif self.edge_case_input(input_dict):
            return False
        return True

    def edge_case_input(self, input_dict: dict[str, Any]) -> bool:
        for line in self.lines:
            matches = expressions.get_matches(line, CHEETAH_EDGE_CASE_INPUT)
            if matches:
                if 'input' not in input_dict:
                    return True
        return False

    def get_eval_strategy(self, input_dict: dict[str, Any]) -> EvaluationStrategy:
        if self.type in [BlockType.INLINE, BlockType.INLINE_ALIAS, BlockType.INLINE_CH]:
            return InlineEvaluationStrategy(self.lines, input_dict)
        elif self.type == BlockType.CONDITIONAL:
            return ConditionalEvaluationStrategy(self.lines, input_dict)
        else:
            raise RuntimeError()








from typing import Any

from galaxy2janis.entities.workflow import Workflow
from galaxy2janis.entities.workflow import WorkflowMetadata


def ingest_metadata(janis: Workflow, galaxy: dict[str, Any]) -> None:
    metadata = parse_metadata(galaxy)
    janis.set_metadata(metadata)

def parse_metadata(galaxy: dict[str, Any]) -> WorkflowMetadata:
    """scrape basic workflow metadata"""
    return WorkflowMetadata(
        name=galaxy['name'],
        uuid=galaxy['uuid'],
        annotation=galaxy['annotation'],
        version=galaxy['version'],
        tags=galaxy['tags']
    )




from bs4 import Tag
from collections import defaultdict
import gx.wrappers.scraping.utils as utils



def scrape_repos() -> None:
    """
    only single thread possible
    returns list of urls, where each url is a repo
    consists of all repos in the toolshed
    """
    URL = 'https://toolshed.g2.bx.psu.edu/repos'
    page = utils.get_soup(URL)
    table = utils.get_table(page)
    elems = utils.get_table_col(table, 0)
    repo_data = extract_repo_information(elems)
    utils.save_data(utils.REPO_DATA_PATH, repo_data)

def extract_repo_information(elems: list[Tag]) -> dict[str, list[str]]:
    out: dict[str, list[str]] = defaultdict(list)
    for elem in elems:
        path = elem.find('a').attrs['href'].split('/') # type: ignore
        owner, repo = path[2], path[3]
        out[owner].append(repo) # type: ignore
    return out







from __future__ import annotations
from .OutputComponent import OutputComponent

from typing import Any, Optional

from ..CommandComponent import CommandComponent


class InputOutput(OutputComponent):
    def __init__(self, input_component: CommandComponent):
        super().__init__()
        self.input_component = input_component
        self.gxparam = self.input_component.gxparam
        assert(self.gxparam)

    @property
    def name(self) -> str:
        name = ''
        if self.gxparam:
            name = self.gxparam.name
        if not name.startswith('out'):
            name = f'out_{name}'
        return name

    @property
    def default_value(self) -> Any:
        raise NotImplementedError()

    @property
    def optional(self) -> bool:
        # NOTE - janis does not allow optional outputs
        return False

    @property
    def docstring(self) -> Optional[str]:
        if self.gxparam:
            return self.gxparam.docstring
        return f'output created during runtime. file relates to the {self.input_component.name} input'

    def update(self, incoming: Any) -> None:
        raise NotImplementedError()







from dataclasses import dataclass
from enum import Enum, auto
from typing import Tuple



class MsgCategory(Enum):
    WARNING     = auto()
    FIX         = auto()
    CHECK       = auto()
    ATTENTION   = auto()


@dataclass
class Message:
    category: MsgCategory
    title: str
    contents: str
    file_reference: str
    line_reference: Tuple[int, int]






from galaxy2janis.gx.gxtool.requirements import Requirement
from ..Container import Container


def select_best_container_match(containers: list[Container], requirement: Requirement) -> Container:
    containers = select_version_matches(containers, requirement)
    containers = sort_by_create_date(containers)
    return containers[0]

def select_version_matches(containers: list[Container], requirement: Requirement) -> list[Container]:
    matches = [c for c in containers if c.tool_version == requirement.version]
    if matches:
        return matches
    # TODO logging.version_mismatch
    return containers

def sort_by_create_date(containers: list[Container]) -> list[Container]: 
    return sorted(containers, key=lambda x: x.timestamp, reverse=True)


# @dataclass
# class SimilarityScore:
#     score: float
#     obj: Any

# def most_similar_version(tool_data: dict[str, Any]) -> dict[str, Any]:
#     vm = VersionMatcher()
#     target_version = requirement.version
#     selected: Optional[dict[str, str]] = None

#     selected = vm.get_version_exact(tool_data, target_version)
#     if not selected:
#         selected = vm.get_version_trimmed(tool_data, target_version)
#     if not selected:
#         selected = vm.get_most_recent(tool_data)
#     return selected
#     # if not selected:
#     #     selected = vm.get_version_trimmed_inexact(query_versions, target_version)

# def select_from_available(images: list[dict[str, Any]]) -> Optional[dict[str, Any]]:
#     for image in images:
#         if image['registry_host'] == 'quay.io/':
#             return image
#     for image in images:
#         if image['registry_host'] == 'depot.galaxyproject.org/singularity/':
#             return image
#     return None



        





from galaxy2janis.logs import logging
import requests
import tarfile
from typing import Optional

from galaxy2janis.utils import galaxy as utils
from galaxy2janis.gx.wrappers.downloads.cache import DownloadCache
from galaxy2janis.gx.interaction import get_builtin_tool_path



CACHE: DownloadCache = DownloadCache()


def fetch_wrapper(owner: str, repo: str, revision: str, tool_id: str) -> str:
    """gets the wrapper locally or from toolshed then returns path to xml"""
    path: Optional[str] = None
    if not path:
        path = _fetch_cache(repo, revision, tool_id)
    if not path:
        path = _fetch_builtin(tool_id)
    if not path:
        path = _fetch_toolshed(owner, repo, revision, tool_id)
    if not path:
        raise RuntimeError(f'could not find wrapper for {tool_id}:{revision}')
    else:
        return path

def _fetch_builtin(tool_id: str) -> Optional[str]:
    return get_builtin_tool_path(tool_id)

def _fetch_cache(repo: str, revision: str, tool_id: str) -> Optional[str]:
    wrapper = CACHE.get(repo, revision)
    if wrapper:
        xml = utils.get_xml_by_id(wrapper, tool_id)
        if xml:
            return f'{wrapper}/{xml}'
    return None

def _fetch_toolshed(owner: str, repo: str, revision: str, tool_id: str) -> Optional[str]:
    # download and add to cache
    url = _get_url_via_revision(owner, repo, revision)
    logging.msg_downloading_tool(url)
    tar = _download_wrapper(url)
    CACHE.add(tar)
    # fetch from cache
    return _fetch_cache(repo, revision, tool_id)

def _get_url_via_revision(owner: str, repo: str, revision: str) -> str:
    return f'https://toolshed.g2.bx.psu.edu/repos/{owner}/{repo}/archive/{revision}.tar.gz'

def _download_wrapper(url: str) -> tarfile.TarFile:
    response = requests.get(url, stream=True)
    return tarfile.open(fileobj=response.raw, mode='r:gz')







from typing import Any
from galaxy2janis.entities.workflow import StepMetadata
from galaxy2janis.gx.wrappers import Wrapper
from galaxy2janis.gx.wrappers import WrapperCache
from galaxy2janis.gx.wrappers.requests.versions import request_installable_revision


"""
This module parses a galaxy step (.ga format) and produces StepMetadata.
During the process it identifies which toolshed Wrapper will be used.

The Wrapper is an important part of StepMetadata. 
It contains information about the tool_id, tool_build, revision,
date_created, conda requirements etc

Local information about Wrappers is stored in data/wrappers/wrappers.json.
If information about which Wrapper to use is not available locally, 
API requests are made to the galaxy toolshed to determine the most suitable wrapper. 
its information is then known and saved to the json file. 

The WrapperCache class is responsible for presenting our local knowledgebank of
wrappers during this phase.
"""


def parse_step_metadata(gxstep: dict[str, Any]) -> StepMetadata:
    return StepMetadata(
        wrapper=get_wrapper(gxstep),
        step_id=int(gxstep['id']),
        step_name=gxstep['name'],
        tool_state=gxstep['tool_state'], 
        workflow_outputs=gxstep['workflow_outputs'],
        _label=gxstep['label']
    )

def get_wrapper(gxstep: dict[str, Any]) -> Wrapper:
    """
    gets a galaxy Wrapper.
    wrapper isn't downloaded - just details are fetched in like owner, repo etc. 
    wrapper download and parsing happens later. 
    wrapper may be pulled from cache of known wrappers, or may a new one may need
    to be created using toolshed API calls. 
    """
    if 'tool_shed_repository' not in gxstep:
        return get_wrapper_builtin(gxstep)
    else:
        return get_wrapper_toolshed(gxstep)

def get_wrapper_builtin(gxstep: dict[str, Any]) -> Wrapper:
    details: dict[str, Any] = {
        'owner': 'Galaxy',
        'repo': 'None',
        'revision': 'None',
        'tool_id': gxstep['tool_id'],
        'tool_build': gxstep['tool_version'],
        'date_created': '1960-01-01 00:00:00',
        'requirements': [],
    }
    wrapper = Wrapper(details)
    wrapper.inbuilt = True
    return wrapper

def get_wrapper_toolshed(gxstep: dict[str, Any]) -> Wrapper:
    wrappers = get_local(gxstep)
    if not wrappers:
        wrappers = get_toolshed(gxstep)
    return most_recent(wrappers)
        
def get_local(gxstep: dict[str, Any]) -> list[Wrapper]:
    cache = WrapperCache()
    return cache.get(
        owner=gxstep['tool_shed_repository']['owner'],
        repo=gxstep['tool_shed_repository']['name'],
        tool_id=gxstep['tool_id'].rsplit('/', 2)[-2],
        tool_build=gxstep['tool_version']
    )

def get_toolshed(gxstep: dict[str, Any]) -> list[Wrapper]:
    # scrape toolshed for wrappers and update cache
    cache = WrapperCache()
    wrappers = request_installable_revision(gxstep)
    for wrapper in wrappers:
        cache.add(wrapper)
    # confirm can now load wrapper details from cache
    local_wrappers = get_local(gxstep)
    if not local_wrappers:
        raise RuntimeError()
    return local_wrappers

def most_recent(wrappers: list[Wrapper]) -> Wrapper:
    """return the most recent wrapper which matches [version]"""
    wrappers.sort(key=lambda x: x.date_created, reverse=True)
    return wrappers[0]






from galaxy2janis.logs import logging
from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional, Tuple
from uuid import uuid4


CH_OPEN_CONDITIONAL = set(['#if ', '#unless '])
CH_WITHIN_CONDITIONAL = set(['#else if ', '#elif ', '#else'])
CH_CLOSE_CONDITIONAL = set(['#end if', '#end unless'])
CH_ALL_CONDITIONAL = CH_OPEN_CONDITIONAL | CH_WITHIN_CONDITIONAL | CH_CLOSE_CONDITIONAL

CH_OPEN_LOOP = set(['#for ', '#while '])
CH_CLOSE_LOOP = set(['#end for', '#end while'])
CH_ALL_LOOP = CH_OPEN_LOOP | CH_CLOSE_LOOP

CH_TRY_EXCEPT = set(['#try ', '#except', '#end try'])

CH_OPEN_FUNC = set(['#def '])
CH_CLOSE_FUNC = set(['#end def'])
CH_ALL_FUNC = CH_OPEN_FUNC | CH_CLOSE_FUNC

CH_ENV = set(['#set ', '#import ', '#from ', '#silent ', '#echo '])

LINUX_ALIAS = set(['set ', 'ln ', 'cp ', 'mv ', 'export '])
LINUX_CMD = set(['mkdir ', 'tar ', 'ls ', 'head ', 'wget ', 'grep ', 'awk ', 'cut ', 'sed ', 'gzip ', 'gunzip ', 'cd ', 'echo ', 'trap ', 'touch '])

CH_CONSTRUCTS = CH_ALL_CONDITIONAL | CH_ALL_LOOP | CH_ALL_FUNC | CH_TRY_EXCEPT | CH_ENV


"""small module which holds <command> section constructs of importance"""


class ConstructType(Enum):
    CONDITIONAL = auto()
    LOOP        = auto()
    FUNCTION    = auto()


@dataclass
class Construct:
    subtype: ConstructType
    
    def __post_init__(self):
        self.lines: list[str] = []
        self.uuid: str = str(uuid4())

# construct subtypes and their opening / closing cheetah tags (for identification in text)
constructs_opens_closes: list[Tuple[ConstructType, set[str], set[str]]] = [
    (ConstructType.CONDITIONAL, CH_OPEN_CONDITIONAL, CH_CLOSE_CONDITIONAL),
    (ConstructType.LOOP, CH_OPEN_LOOP, CH_CLOSE_LOOP),
    (ConstructType.FUNCTION, CH_OPEN_FUNC, CH_CLOSE_FUNC),
]


class ConstructStack:
    def __init__(self):
        self.stack: list[Construct] = []
    
    def add(self, construct: Construct) -> None:
        self.stack.append(construct)
        if construct.subtype == ConstructType.LOOP:
            logging.has_cheetah_loop()
        elif construct.subtype == ConstructType.FUNCTION:
            logging.has_cheetah_function()
    
    def pop(self, subtype: ConstructType) -> None:
        if self.depth > 0:
            assert(self.current_construct)
            assert(self.current_construct.subtype == subtype)
            self.stack.pop()

    @property
    def depth(self) -> int:
        return len(self.stack)
    
    @property
    def current_construct(self) -> Optional[Construct]:
        if len(self.stack) > 0:
            return self.stack[-1]
        return None

    def within(self, ctype: ConstructType) -> bool:
        for construct in self.stack:
            if construct.subtype == ctype:
                return True
        return False


class ConstructTracker:

    def __init__(self):
        self.stack: ConstructStack = ConstructStack()
    
    # methods to manage construct stack
    def update(self, line: str):
        self.handle_entering_constructs(line)
        self.handle_line(line)
        self.handle_leaving_constructs(line)

    def handle_entering_constructs(self, line: str) -> None:
        for subtype, openings, _ in constructs_opens_closes:
            if any ([line.startswith(openstr) for openstr in openings]):
                construct = Construct(subtype)
                self.stack.add(construct)

    def handle_line(self, line: str) -> None:
        if self.stack.current_construct: 
            self.stack.current_construct.lines.append(line)

    def handle_leaving_constructs(self, line: str) -> None:
        for subtype, _, closings in constructs_opens_closes:
            if any ([line.startswith(closestr) for closestr in closings]):
                self.stack.pop(subtype)

    @property
    def within_conditional(self) -> bool:
        if self.stack.within(ConstructType.CONDITIONAL):
            return True
        return False
    
    @property
    def within_loop(self) -> bool:
        if self.stack.within(ConstructType.LOOP):
            return True
        return False
    
    @property
    def within_banned_segment(self) -> bool:
        banned_constructs = [ConstructType.LOOP, ConstructType.FUNCTION]
        for construct in banned_constructs: 
            if self.stack.within(construct):
                return True
        return False

    def active_is_boundary(self, line: str) -> bool:
        if any ([line.startswith(kw) for kw in CH_CONSTRUCTS]):
            return True
        return False

